#pragma once

#include "layers/common/include/cudnn_frontend_layer_interface.h"

#include "layers/models/resnet/forward/cudnn_frontend_residual_forward_block.h"
#include "layers/models/resnet/forward/cudnn_frontend_residual_forward_inference_block.h"